/*
 * Copyright (C) 2012 Marcus Overhagen <marcus@overhagen.de>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * version 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301, USA.
 */
#include "LEDStrip.h"
#include "spi.h"
#include "util.h"


#include <string.h>

LEDStrip::LEDStrip(uint16_t ledCount)
 :	fLedCount(ledCount),
  	fBufferSize(ledCount * 3)
{
	fBuffer = new uint8_t [fBufferSize];
	memset(fBuffer, 0, fBufferSize);
	_Show();
}


LEDStrip::~LEDStrip()
{
}

void
LEDStrip::_Show()
{
	spi_write_lpd8806(fBuffer, fBufferSize);
}


void
LEDStrip::ColorChase(rgb_t background, rgb_t color, uint16_t delayTime, uint16_t repeatCount)
{
	_Fill(background);
	while (repeatCount--) {
		for (uint16_t i = 0; i <= End(); i++) {
			_Set(i == 0 ? End() : i - 1, background);
			_Set(i, color);
			_Show();
			delay(delayTime);
		}
	}
}


void
LEDStrip::ColorChaseReverse(rgb_t background, rgb_t color, uint16_t delayTime, uint16_t repeatCount)
{
	_Fill(background);
	while (repeatCount--) {
		for (int16_t i = End() ; i >= 0; i--) {
			_Set(i == End() ? 0 : i + 1, background);
			_Set(i, color);
			_Show();
			delay(delayTime);
		}
	}
}


void
LEDStrip::BlockColorChase(rgb_t background, rgb_t color, uint16_t blockLength, uint16_t distanceLength, uint16_t delayTime, uint16_t repeatCount)
{
	/*
	while (repeatCount--) {
		for (uint16_t k = 0; k < Count(); k++) {
			for (uint16_t i = 0; ; ) {
				for (uint16_t j = 0; j < blockLength && i <= End(); i++, j++)
					_Set(i, color);
				for (uint16_t j = 0; j < distanceLength && i <= End(); i++, j++)
					_Set(i, background);
			}
			_Show();
			delay(delayTime);
		}
	}
	*/
}


void
LEDStrip::BlockColorChaseReverse(rgb_t background, rgb_t color, uint16_t blockLength, uint16_t distanceLength, uint16_t delay, uint16_t repeatCount)
{
}


void
LEDStrip::ColorFill(rgb_t color, uint16_t delayTime)
{
	_Fill(color);
	_Show();
	delay(delayTime);
}


void
LEDStrip::Wipe(rgb_t color, uint16_t delayTime)
{
	for (uint16_t i = 0; i <= End(); i++) {
		_Set(i, color);
		_Show();
		delay(delayTime);
	}
}


void
LEDStrip::WipeReverse(rgb_t color, uint16_t delayTime)
{
	for (int16_t i = End() ; i >= 0; i--) {
		_Set(i, color);
		_Show();
		delay(delayTime);
	}
}


// Input value 0 to 767 to get a color value from RGB wheel
// Colours are a transition r - g - b - r
rgb_t
LEDStrip::rgb_wheel(uint16_t index)
{
	uint8_t r;
	uint8_t g;
	uint8_t b;
	uint8_t index8bit;

	index8bit = (uint8_t)index;

	switch (index >> 8)
	{
		case 0:
			r = 0xff - index8bit;   // red down
			g = index8bit;          // green up
			b = 0;                  // blue off
			break;
		case 1:
			r = 0;                  // red off
			g = 0xff - index8bit;   // green down
			b = index8bit;          // blue up
			break;
		case 2:
			r = index8bit;          // red up
    		g = 0;                  // green off
    		b = 0xff - index8bit;   // blue down
    		break;
	}

	r = (r * (RGB_RED_MAX + 1)) >> 8;
	g = (g * (RGB_GREEN_MAX + 1)) >> 8;
	b = (b * (RGB_BLUE_MAX + 1)) >> 8;

/*
	uint16_t temp = r + r;
	if (temp > 255)
		r = 255;
	else
		r = temp;
*/
	return rgb(r,g,b);
}


void
LEDStrip::Rainbow(uint16_t steps, uint16_t delayTime, uint16_t repeatCount)
{
	if (steps > 768)
		steps = 768;

	while (repeatCount--) {
		for (int16_t j = 2*steps - 1; j >= 0; j--) {
			for (uint16_t i=0; i < Count(); i++) {
				_Set(i, rgb_wheel(((i + j) * (uint32_t)768 / steps)  % 768));

			}
	    	_Show();
			delay(delayTime);
		}
	}
}

void
LEDStrip::RainbowReverse(uint16_t steps, uint16_t delayTime, uint16_t repeatCount)
{
	if (steps > 768)
		steps = 768;

	while (repeatCount--) {
		for (uint16_t j = 0; j < steps *2; j++) {
			for (uint16_t i = 0; i < Count(); i++) {
				_Set(i, rgb_wheel(((i + j) * (uint32_t)768 / steps)  % 768));
			}
			_Show();
			delay(delayTime);
		}
	}
}


void
LEDStrip::Cycle(uint16_t steps, uint16_t delayTime, uint16_t repeatCount)
{
	if (steps > 768)
		steps = 768;

	while (repeatCount--) {
		for (uint16_t i = 0; i < steps; i++) {
			_Fill(rgb_wheel(i * (uint32_t)768 / steps));
			_Show();
			delay(delayTime);
		}
	}
}


void
LEDStrip::CycleReverse(uint16_t steps, uint16_t delayTime, uint16_t repeatCount)
{
	if (steps > 768)
		steps = 768;

	while (repeatCount--) {
		for (int16_t i = steps - 1; i >= 0; i--) {
			_Fill(rgb_wheel(i * (uint32_t)768 / steps));
			_Show();
			delay(delayTime);
		}
	}
}
/*
const uint8_t exptab[32] =
{
	1,	1,	1,	2,
	2,	2,	3,	3,
	4,	5,	6,	7,
	9,	10,	12,	15,
	17,	21,	25,	30,
	36,	43,	51,	61,
	73,	87,	104, 125,
	149, 178, 213, 255
};
*/

// Anfangswerte geändert und nicht mehr ganz so hübsch, aber vermeidet das
// lange Leuchten einzelner farben wenn andere noch aus sind.
// Berechnet mit http://www.mikrocontroller.net/wikifiles/e/e4/Pwm_table.zip
const uint8_t exptab[32] =
{
	0,		5,		10,		15,
	20,		24,		26,		28,
	31,		34,		37,		41,
	45,		49,		54,		59,
	65,		71,		78,		85,
	93,		102,	112,	123,
	134,	147,	161,	177,
	194,	212,	233,	255
};


void
LEDStrip::FadeIn(rgb_t color, uint16_t delayTime)
{
	uint16_t rIn = rgb_r(color);
	uint16_t gIn = rgb_g(color);
	uint16_t bIn = rgb_b(color);

	for (uint8_t i = 0; i < 32; i++)
	{
		uint8_t r = (rIn * exptab[i]) / 256;
		uint8_t g = (gIn * exptab[i]) / 256;
		uint8_t b = (bIn * exptab[i]) / 256;
		_Fill(r, g, b);
		_Show();
		delay(delayTime);
	}
}


void
LEDStrip::FadeOut(rgb_t color, uint16_t delayTime)
{
	uint16_t rIn = rgb_r(color);
	uint16_t gIn = rgb_g(color);
	uint16_t bIn = rgb_b(color);

	for (int8_t i = 31; i >= 0; i--)
	{
		uint8_t r = (rIn * exptab[i]) / 256;
		uint8_t g = (gIn * exptab[i]) / 256;
		uint8_t b = (bIn * exptab[i]) / 256;
		_Fill(r, g, b);
		_Show();
		delay(delayTime);
	}
}




rgb_t reduce(rgb_t color, uint8_t i)
{
	int16_t rIn = rgb_r(color);
	int16_t gIn = rgb_g(color);
	int16_t bIn = rgb_b(color);
	uint8_t r = (rIn * exptab[i]) >> 8;
	uint8_t g = (gIn * exptab[i]) >> 8;
	uint8_t b = (bIn * exptab[i]) >> 8;
	return rgb(r,g,b);
}


rgb_t dimm(rgb_t color, int16_t fraction)
{
	uint8_t rIn = rgb_r(color);
	uint8_t gIn = rgb_g(color);
	uint8_t bIn = rgb_b(color);
	uint8_t r = (rIn * fraction) >> 8;
	uint8_t g = (gIn * fraction) >> 8;
	uint8_t b = (bIn * fraction) >> 8;
	return rgb(r,g,b);
}


void
LEDStrip::KnightRider(rgb_t background, rgb_t color, uint16_t startPos, uint8_t length, uint16_t delayTime, uint16_t repeatCount)
{
	_Fill(background);
	uint16_t endPos = startPos + length -1;
	while (repeatCount--) {
//		for (uint8_t j = 0; j < length; j++) {
			for (uint16_t i = startPos; i <= endPos; i++) {
				_SetChecked(startPos, endPos, i - 5, background);
				_SetChecked(startPos, endPos, i - 4, dimm(color, 4));
				_SetChecked(startPos, endPos, i - 3, dimm(color, 22));
				_SetChecked(startPos, endPos, i - 2, dimm(color, 40));
				_SetChecked(startPos, endPos, i - 1, dimm(color, 55));
				_Set(i, color);
/*				
				_SetChecked(startPos, endPos, i + 1, dimm(color, 55));
				_SetChecked(startPos, endPos, i + 2, dimm(color, 40));
				_SetChecked(startPos, endPos, i + 3, dimm(color, 22));
				_SetChecked(startPos, endPos, i + 4, dimm(color, 4));
*/				
				_Show();
				delay(delayTime);
/*				
				if (i == startPos || i == endPos)
					delay(delayTime / 3);
*/				
			}
			for (uint16_t i = endPos - 1; i > startPos; i--) {
/*				
				_SetChecked(startPos, endPos, i - 4, dimm(color, 4));
				_SetChecked(startPos, endPos, i - 3, dimm(color, 22));
				_SetChecked(startPos, endPos, i - 2, dimm(color, 40));
				_SetChecked(startPos, endPos, i - 1, dimm(color, 55));
*/				
				_Set(i, color);
				_SetChecked(startPos, endPos, i + 1, dimm(color, 55));
				_SetChecked(startPos, endPos, i + 2, dimm(color, 40));
				_SetChecked(startPos, endPos, i + 3, dimm(color, 22));
				_SetChecked(startPos, endPos, i + 4, dimm(color, 4));
				_SetChecked(startPos, endPos, i + 5, background);
				_Show();
				delay(delayTime);
			}
//		}
	}
}
