/*
 * Copyright (C) 2012 Marcus Overhagen <marcus@overhagen.de>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * version 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301, USA.
 */
#ifndef _LEDSTRIP_H
#define _LEDSTRIP_H

#include <inttypes.h>
#include <stdio.h>
#include <stdlib.h>
#include <avr/io.h>
#include <avr/interrupt.h>

#if 1

// 2568 bytes

typedef uint32_t rgb_t;

inline rgb_t
rgb(uint8_t r, uint8_t g, uint8_t b)
{
	return (uint32_t)r << 16 | (uint16_t)g << 8 | b;
}

#define rgb_r(c)	((uint8_t)((c) >> 16))
#define rgb_g(c)	((uint8_t)((c) >> 8))
#define rgb_b(c)	((uint8_t)(c))

#else

// 2776 bytes

typedef struct
{
	uint8_t r;
	uint8_t g;
	uint8_t b;
} rgb_t;


inline rgb_t
rgb(uint8_t r, uint8_t g, uint8_t b)
{
	rgb_t rgb;
	rgb.r = r;
	rgb.g = g;
	rgb.b = b;
	return rgb;
}

#define rgb_r(c)	(c).r
#define rgb_g(c)	(c).g
#define rgb_b(c)	(c).b

#endif

#if 1
	#define RGB_RED_MAX		255
	#define RGB_GREEN_MAX	255
	#define RGB_BLUE_MAX	255

	#define black		rgb(0x00,0x00,0x00)
	#define white		rgb(0xff,0xff,0xff)

	#define red			rgb(0xff,0x00,0x00)
	#define green		rgb(0x00,0xff,0x00)
	#define blue		rgb(0x00,0x00,0xff)

	#define yellow		rgb(0xff,0xff,0x00)
	#define cyan		rgb(0x00,0xff,0xff)
	#define magenta		rgb(0xff,0x00,0xff)

	#define orange		rgb(0xff,0x88,0x00)

#else
	#define RGB_RED_MAX		255
	#define RGB_GREEN_MAX	120
	#define RGB_BLUE_MAX	60

	#define black		rgb(0x00,0x00,0x00)
	#define white		rgb(RGB_RED_MAX,RGB_GREEN_MAX,RGB_BLUE_MAX)

	#define red			rgb(0xff,0x00,0x00)
	#define green		rgb(0x00,0xff,0x00)
	#define blue		rgb(0x00,0x00,0xff)

	#define yellow		rgb(RGB_RED_MAX,RGB_GREEN_MAX,0x00)
	#define cyan		rgb(0x00,RGB_GREEN_MAX,RGB_BLUE_MAX)
	#define magenta		rgb(RGB_RED_MAX,0x00,RGB_BLUE_MAX)

	#define orange		rgb(RGB_RED_MAX,(0x88 * (RGB_GREEN_MAX + 1)) >> 8,0x00)

#endif


class LEDStrip
{
public:
				LEDStrip(uint16_t ledCount);
				~LEDStrip();

	void		ColorFill(rgb_t color, uint16_t delay);


	void		Wipe(rgb_t color, uint16_t delay);
	void		WipeReverse(rgb_t color, uint16_t delay);

	void		ColorChase(rgb_t background, rgb_t color, uint16_t delay, uint16_t repeatCount = 1);
	void		ColorChaseReverse(rgb_t background, rgb_t color, uint16_t delay, uint16_t repeatCount = 1);

	void		BlockColorChase(rgb_t background, rgb_t color, uint16_t blockLength, uint16_t distanceLength, uint16_t delay, uint16_t repeatCount = 1);
	void		BlockColorChaseReverse(rgb_t background, rgb_t color, uint16_t blockLength, uint16_t distanceLength, uint16_t delay, uint16_t repeatCount = 1);

	void		Rainbow(uint16_t steps, uint16_t delay, uint16_t repeatCount = 1);
	void		RainbowReverse(uint16_t steps, uint16_t delay, uint16_t repeatCount = 1);

	void		Cycle(uint16_t steps, uint16_t delay, uint16_t repeatCount = 1);
	void		CycleReverse(uint16_t steps, uint16_t delay, uint16_t repeatCount = 1);

	void		FadeIn(rgb_t color, uint16_t delay);
	void		FadeOut(rgb_t color, uint16_t delay);


	void		KnightRider(rgb_t background, rgb_t color, uint16_t startPos, uint8_t length, uint16_t delay, uint16_t repeatCount = 20);

	uint16_t	Count();
	uint16_t	End();

private:
	void		_Show();
	void		_Set(uint16_t led, rgb_t color);
	void		_Set(uint16_t led, uint8_t r, uint8_t g, uint8_t b);
	void		_SetChecked(uint16_t startPos, uint16_t endPos, uint16_t led, rgb_t color);
	void		_Fill(rgb_t color);
	void		_Fill(uint8_t r, uint8_t g, uint8_t b);

static rgb_t	rgb_wheel(uint16_t index);

private:
	uint16_t	fLedCount;
	uint16_t	fBufferSize;
	uint8_t *	fBuffer;
};


inline uint16_t
LEDStrip::Count()
{
	return fLedCount;
}


inline uint16_t
LEDStrip::End()
{
	return fLedCount - 1;
}


inline void
LEDStrip::_Set(uint16_t led, uint8_t r, uint8_t g, uint8_t b)
{
//	if (r < 127) r *= 2; else { g /= 3; b /= 3; }

	if (led < fLedCount) {
		uint8_t *buf = fBuffer + led * 3;
		buf[0] = r;
		buf[1] = g;
		buf[2] = b;
	}
}


inline void
LEDStrip::_Set(uint16_t led, rgb_t color)
{
	_Set(led, rgb_r(color), rgb_g(color), rgb_b(color));
}


inline void
LEDStrip::_SetChecked(uint16_t startPos, uint16_t endPos, uint16_t led, rgb_t color)
{
	if (led >= startPos && led <= endPos)
		_Set(led, color);
}


inline void
LEDStrip::_Fill(uint8_t r, uint8_t g, uint8_t b)
{
//	if (r < 127) r *= 2; else { g /= 3; b /= 3; }

	uint8_t *buf = fBuffer;
	for (uint16_t i = 0; i < fLedCount; i++) {
		buf[0] = r;
		buf[1] = g;
		buf[2] = b;
		buf += 3;
	}
}


inline void
LEDStrip::_Fill(rgb_t color)
{
	_Fill(rgb_r(color), rgb_g(color), rgb_b(color));
}


#endif
